#
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004, JSC Intervale
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#	this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#	this list of conditions and the following disclaimer in the documentation
#	and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#	from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

require 'mxx_ru/binary_target'
require 'mxx_ru/cpp/mode'
require 'mxx_ru/cpp/source_file'
require 'mxx_ru/cpp/obj_placement'
require 'mxx_ru/cpp/analyzer'

#      /C++ .
module Mxx_ru

	module Cpp
		#  runtime_mode.

		#  -.
		#       ,
		#    .
		RUNTIME_DEFAULT = "default"
		# Release-.  .
		RUNTIME_RELEASE = "release"
		# Debug-.   .
		RUNTIME_DEBUG = "debug"

		#  rtti_mode.

		#  rtti -.   
		#    ,     RTTI.
		RTTI_DEFAULT = "default"
		#  RTTI   .
		RTTI_ENABLED = "enabled"
		#  RTTI   .
		RTTI_DISABLED = "disabled"

		#  rtl_mode.

		#  ,   
		#  .
		RTL_DEFAULT = "default"
		#   runtime- .
		RTL_SHARED = "shared"
		#   runtime- .
		RTL_STATIC = "static"

		#  threading_mode.

		#  ,   
		# -.
		THREADING_DEFAULT = "default"
		#  .
		THREADING_MULTI = "multi"
		#   .
		THREADING_SINGLE = "single"

		#  screen_mode.

		#    .
		SCREEN_CONSOLE = "console"
		#    .
		SCREEN_WINDOW = "window"

		#  .

		#     .
		OPTIM_SIZE = "size"
		#   .
		OPTIM_SPEED = "speed"

		#  ,   
		#    runtime_mode, rtl_mode, ..., 
		#   ,      .
		class	Global_single_value_option
			attr_reader	:name
			attr_reader	:value
			attr_reader	:who

			# _a_name_ .
			#
			# _a_default_value_  -.
			#
			# _a_available_values_  . Array  String.
			#
			def initialize(
				a_name,
				a_default_value,
				a_available_values )

				@name = a_name
				@default_value = a_default_value
				@available_values = a_available_values
				@value = @default_value
				@who = "n/a"
			end

			#  .
			#      -, 
			#    .
			#
			# _a_value_  .
			#
			# _a_who_    .
			#
			def change(
				a_value,
				a_who )

				if a_value != @default_value
					if @available_values.include?( a_value )
						if @value == @default_value
							#      .
							@value = a_value
							@who = a_who
						else
							#   .  , 
							#  .     .
							if a_value != @value
								raise Mxx_ru::Global_option_conflict_ex.new(
									@name, @value, @who, a_value, a_who )
							end
						end
					else
						raise Mxx_ru::Global_option_disallowed_value_ex.new(
							@name, @available_values, a_value, a_who )
					end
				end
			end
		end # class Global_single_value_option

		#     ,  
		#    3 : ,  
		#  (up-spread).
		#    : defines, include_paths,
		# compiler_options, linker_options, ...
		class	Spreadable_option
			#  ,    
			#   .
			ALL_OPT_METHOD = "all"
			#  ,    
			#    .
			ALL_UPSPREAD_METHOD = "all_upspreads"

			#       
			# ,     .
			#      add.
			@@last_change_time = 1
			#  ,     
			#     .
			@@cache_times = Hash.new

			#      .
			#       
			#  ,    
			# spreadable-,    
			#   spreadable-.
			def	Spreadable_option.refresh_option_change_time( a_option_name )
				@@last_change_time = @@last_change_time + 1
				@@cache_times[ a_option_name ] = @@last_change_time
			end	

			# .
			#
			# _a_option_name_
			#  ,    .
			#        
			#   @@cache_times.
			def	initialize( a_option_name )

				@@cache_times[ a_option_name ] = 1

				#  .
				@name = a_option_name

				#  .
				@locals = Array.new
				#  .
				@upspreads = Array.new

				#    all.
				@all_cache_value = Array.new
				#  @@last_change_time      all.
				@all_cache_time = 0

				#    all_upspread.
				@all_upspread_cache_value = Array.new
				#  @@last_change_time      all_upspread.
				@all_upspread_cache_time = 0

			end

			#  .
			#
			# _a_option_  .
			#
			# _a_mode_    .
			#   Mxx_ru::Cpp::Target::OPT_LOCAL,
			# Mxx_ru::Cpp::Target::OPT_UPSPEAD.
			def	add(
				a_option,
				a_mode = Mxx_ru::Cpp::Target::OPT_LOCAL )
				add_unique_to( a_option, check_opt_mode( a_mode ) )
			end

			#    .
			def	upspeads_only
				return @upspreads
			end

			#    .
			#
			# _a_globals_   ,    
			#   .
			#
			# _a_subprojects_
			#    .
			# Array  Mxx_ru::Abstract_target.
			#
			# _a_upspreads_method_
			#  ,      ,
			#        .
			def	all(
				a_globals,
				a_subprojects,
				a_upspreads_method )

				if @all_cache_time < @@cache_times[ @name ]
					#    .
					r = Array.new
					r << a_globals << @locals

					#     upspread-.
					if @all_upspread_cache_time < @@cache_times[ @name ]
						u = Array.new
						u << @upspreads
            
						a_subprojects.each { |p|
							if p.kind_of?( Mxx_ru::Cpp::Target )
								u << p.send( a_upspreads_method )
							end
						}

						@all_upspread_cache_time = @@last_change_time
						@all_upspread_cache_value = u.flatten.uniq
					end

					r << @all_upspread_cache_value
            
					@all_cache_value = r.flatten.uniq
					@all_cache_time = @@last_change_time
				end

				return @all_cache_value
			end

			#     .
			#
			# _a_subprojects_
			#    .
			# Array  Mxx_ru::Abstract_target.
			#
			# _a_upspreads_method_
			# ,      ,
			#        .
			def	all_upspreads(
				a_subprojects,
				a_upspreads_method )

				if @all_upspread_cache_time < @@cache_times[ @name ]
					#    .
					r = Array.new
					r << @upspreads
   
					a_subprojects.each { |p|
						if p.kind_of?( Mxx_ru::Cpp::Target )
							r << p.send( a_upspreads_method )
						end
					}
   
					@all_upspread_cache_value = r.flatten.uniq
					@all_upspread_cache_time = @@last_change_time
				end

				return @all_upspread_cache_value
			end

		protected
			#    Array,    
			# Array  .
			def add_unique_to( what, to )
				if !to.include?( what )
					to.push( what )

					Spreadable_option::refresh_option_change_time( @name )
				end
			end

			#    mode   
			# OPT_UPSPREAD,  OPT_LOCAL   
			#  Array,     .
			#
			# _a_mode_
			#   mode  .
			def	check_opt_mode(
				a_mode )
				if Mxx_ru::Cpp::Target::OPT_UPSPREAD == a_mode
					return @upspreads
				elsif Mxx_ru::Cpp::Target::OPT_LOCAL == a_mode
					return @locals
				else
					raise Mxx_ru::Invalid_value_ex(
						"invalid option mode #{a_mode}" )
				end
			end
		end

		#   (exe, lib, dll, unittest, composite  ..).
		#      .
		class	Target_type
			def name
				raise Abstract_method_ex.new( "Mxx_ru::Cpp::Target_type::name" )
			end
		end

		#    exe-.
		class	Exe_target_type < Target_type
			TYPE = "exe"

			def name
				return TYPE
			end
		end

		#      (lib).
		class	Lib_target_type < Target_type
			TYPE = "lib"

			def name
				return TYPE
			end
		end

		#    -  (dll).
		class	Dll_target_type < Target_type
			TYPE = "dll"

			def name
				return TYPE
			end
		end

		#      C/C++ .
		class	Target < Mxx_ru::Binary_target
			#    obj_placement.
			#     obj_placement.
			class	Global_obj_placement_info
				#  obj_placement.
				attr_reader	:placement
				#   obj_placement.
				attr_reader	:who

				def initialize
					@placement = nil
					@who = nil
				end

				#   .
				#  placement  ,   .
				def	change( a_placement, a_who )
					if nil != a_placement && nil != @placement
						raise Mxx_ru::Global_option_conflict_ex.new(
							"global_obj_placement",
							@placement.class.name,
							@who,
							a_placement.class.name,
							a_who )
					end

					@placement = a_placement
					@who = a_who
				end
			end

			#       Array.
			class	Global_option_inserter

				# _a_option_name_
				#   .     
				# Spreadable_option.refresh_option_change_time.
				#
				# _a_receiver_
				#    .
				def	initialize(
					a_option_name,
					a_receiver )
					@option_name = a_option_name
					@receiver = a_receiver
				end

				def	call( *args )
					@receiver.push( *args )

					Spreadable_option::refresh_option_change_time( @option_name )
				end
			end # class Global_option_inserter

			#        Spreadable_option.
			class	Spreadable_option_inserter

				# _a_receiver_
				#    .
				def	initialize(
					a_receiver )
					@receiver = a_receiver
				end

				def	call( *args )
					@receiver.send( "add", *args )
				end
			end # class Spreadable_option_inserter

			#       Spreadable_option
			#    (,   ).
			class	Spreadable_option_all_extractor

				# _a_option_
				#     call.
				#
				# _a_globals_
				#     .
				#
				# _a_target_
				# ,       .
				#
				# _a_upspreads_only_method_
				# ,      
				#      .
				def	initialize(
					a_option,
					a_globals,
					a_target,
					a_upspreads_only_method )

					@option = a_option
					@globals = a_globals
					@target = a_target
					@upspreads_only_method = a_upspreads_only_method
				end

				def call( *args )
					return @option.all( @globals, @target.mxx_required_prjs,
						@upspreads_only_method )
				end
			end # class Spreadable_option_all_extractor

			#       Spreadable_option
			#    .
			class	Spreadable_option_upspread_extractor

				# _a_option_
				#     call.
				#
				# _a_target_
				# ,       .
				#
				# _a_upspreads_only_method_
				# ,      
				#      .
				def	initialize(
					a_option,
					a_target,
					a_upspreads_only_method )

					@option = a_option
					@target = a_target
					@upspreads_only_method = a_upspreads_only_method
				end

				def call( *args )
					return @option.all_upspreads( @target.mxx_required_prjs,
						@upspreads_only_method )
				end
			end # class Spreadable_option_all_extractor

			#   runtime.
			@@mxx_runtime_mode = Global_single_value_option.new(
				"runtime_mode", RUNTIME_DEFAULT,
				[ RUNTIME_RELEASE, RUNTIME_DEBUG ] )
			#   RTTI.
			@@mxx_rtti_mode = Global_single_value_option.new(
				"rtti_mode", RTTI_DEFAULT,
				[ RTTI_ENABLED, RTTI_DISABLED ] )
			#   runtime-.
			@@mxx_rtl_mode = Global_single_value_option.new(
				"rtl_mode", RTL_DEFAULT,
				[ RTL_SHARED, RTL_STATIC ] )
			#   .
			@@mxx_threading_mode = Global_single_value_option.new(
				"threading_mode", THREADING_DEFAULT,
				[ THREADING_MULTI, THREADING_SINGLE ] )

			#    obj_placement.
			@@mxx_global_obj_placement = Global_obj_placement_info.new

			#   .
			#   String,  -- Array  String.
			@@mxx_globals = Hash.new

			#  ,     .
			OPT_UPSPREAD = "upspread"
			#  ,     .
			OPT_LOCAL = "local"

			# ,      .
			#    ,    
			#  .     mxx_target_root
			#     .
			attr_reader	:mxx_target_root
			# ,      .
			attr_reader	:mxx_sources_root
			#  . -, .
			attr_reader	:mxx_screen_mode
			#   C-.
			#   Mxx_ru::Cpp::Cpp_source_file.
			attr_reader	:mxx_c_files
			#   C++-.
			#   Mxx_ru::Cpp::Cpp_source_file.
			attr_reader	:mxx_cpp_files
			#   .
			attr_reader	:mxx_obj_files
			#  mswin- .
			#     Mxx-ru::Cpp::Mswin_rc_file.
			#    nil,  mswin-  .
			attr_reader	:mxx_mswin_rc_file
			#  mswin-  .
			#   .
			#    nil,  mswin-  .
			attr_reader	:mxx_mswin_res_file
			#    .
			#    nil.
			attr_reader	:mxx_implib_path
			#     .
			#    .
			attr_reader	:mxx_target_name
			#  . -  OPTIM_SPEED.
			attr_reader	:mxx_optimization

			# .
			# a_prj_alias -  .
			def initialize( a_prj_alias )

				super( a_prj_alias )

				@mxx_target_root = ""

				@mxx_sources_root = File.dirname( a_prj_alias )

				@mxx_screen_mode = SCREEN_CONSOLE

				@mxx_c_files = Array.new
				@mxx_cpp_files = Array.new

				@mxx_mswin_rc_file = nil
				@mxx_mswin_res_file = nil

				@mxx_obj_files = Array.new

				#   .
				#   String,  -- Spreadable_option.
				@mxx_locals = Hash.new

				@mxx_implib_path = nil

				# ,     .
				@mxx_obj_placement = nil

				@mxx_target_name = nil

				@mxx_optimization = OPTIM_SPEED

				#       build.
				#    build    ,
				#     .
				@mxx_last_build_result = nil

				#  ,  toolset    
				#     .
				@mxx_all_options_defined = false

				#    ,  
				#    method_missing.
				@mxx_missing_actions = Hash.new

				add_spreadable_option( "compiler_option" )
				add_spreadable_option( "cpp_compiler_option" )
				add_spreadable_option( "c_compiler_option" )
				add_spreadable_option( "linker_option" )
				add_spreadable_option( "librarian_option" )
				add_spreadable_option( "define" )
				add_spreadable_option( "include_path" )
				add_spreadable_option( "mswin_rc_option" )
				add_spreadable_option( "mswin_rc_include_path" )
				add_spreadable_option( "mswin_rc_define" )
				add_spreadable_option( "mswin_rlink_option" )

			end

			#      
			#  .       .
			#       
			#  @mxx_missing_actions.
			def	method_missing( a_meth_id, *args )
				meth_name = a_meth_id.id2name
				caller = @mxx_missing_actions[ meth_name ]
				if caller
					r = caller.call( *args )
				else
					super( a_meth_id, *args )
				end
			end

			#      toolset.
			#  ,   Mxx_ru::Cpp::Target 
			#   toolset,   Mxx_ru::Cpp::toolset.
			#    Mxx_ru::Cpp::toolset.
			def	toolset()
				return Mxx_ru::Cpp::toolset()
			end

			#   runtime_mode.
			def	runtime_mode( mode )
				@@mxx_runtime_mode.change( mode, prj_alias )
			end
			#   runtime_mode.
			def	mxx_runtime_mode
				#  runtime_mode   RUNTIME_DEFAULT (..
				#    ,    
				#    (--mxx-cpp-release, --mxx-cpp-debug).
				result = @@mxx_runtime_mode.value
				if RUNTIME_DEFAULT == result
					if Mxx_ru::Cpp::Mode.instance.is_release &&
						Mxx_ru::Cpp::Mode.instance.is_debug
						#    .
						raise Mxx_ru::Invalid_value_ex.new(
							"#{Mxx_ru::Cpp::MXXCPPARG_RELEASE} and " +
							"#{Mxx_ru::Cpp::MXXCPPARG_DEBUG} cannot be used " +
							"together" )
					elsif Mxx_ru::Cpp::Mode.instance.is_release
						result = RUNTIME_RELEASE
					elsif Mxx_ru::Cpp::Mode.instance.is_debug
						result = RUNTIME_DEBUG
					end
				end

				return result
			end

			#   rtti_mode.
			def	rtti_mode( mode )
				@@mxx_rtti_mode.change( mode, prj_alias )
			end
			#   rtti_mode.
			def	mxx_rtti_mode
				return @@mxx_rtti_mode.value
			end

			#   rtl_mode.
			def	rtl_mode( mode )
				@@mxx_rtl_mode.change( mode, prj_alias )
			end
			#   rtl_mode.
			def	mxx_rtl_mode
				return @@mxx_rtl_mode.value
			end

			#   threading_mode.
			def	threading_mode( mode )
				@@mxx_threading_mode.change( mode, prj_alias )
			end
			#   threading_mode.
			def	mxx_threading_mode
				return @@mxx_threading_mode.value
			end

			#   obj_placement.
			#   obj_placement  , 
			#  .
			def	global_obj_placement( a_placement )
				@@mxx_global_obj_placement.change( a_placement, prj_alias )
			end

			#   obj_placement.
			#      
			#  obj_placement.
			def	obj_placement( a_placement )
				@mxx_obj_placement = a_placement
			end

			#   obj_placement,   .
			#       obj_placement,
			#   .
			#    ,   obj_placement,
			#   .
			#     ,   obj_placement,
			#    obj_placement   
			# Mxx_ru::Cpp::default_obj_placement    .
			def	mxx_obj_placement
				if nil != @mxx_obj_placement &&
					nil != @@mxx_global_obj_placement.placement
					raise Mxx_ru::Invalid_value_ex.new(
						"Global and local obj_placement object exists! " +
						"Global created by: #{@@mxx_global_obj_placement.who}. " +
						"Local created by: #{prj_alias}" )
				end

				if nil != @@mxx_global_obj_placement.placement
					return @@mxx_global_obj_placement.placement
				end

				if nil == @mxx_obj_placement
					@mxx_obj_placement = Mxx_ru::Cpp::default_obj_placement
				end

				return @mxx_obj_placement
			end

			#  ,    
			#  .
			#        target!
			#  target_root     target,
			#   _a_root_    
			# .
			def	target_root( a_root )
				@mxx_target_root = a_root
				if @mxx_target_name
					@mxx_target_name = File.join(
						[ @mxx_target_root, @mxx_target_name ] )
				end
			end

			#        .
			def	target( a_target )
				@mxx_target_name = create_full_target_file_name( a_target )
			end

			#  ,    
			#  .    ,
			#       .
			#   ,     ,  ,
			#      .
			#
			#    mxx_sources_root. 
			#  ,     -.
			def	sources_root( a_root )
				old_root = @mxx_sources_root.clone
				if block_given?
					@mxx_sources_root = File.join( [ @mxx_sources_root, a_root ] )
					yield
					@mxx_sources_root = old_root
				else
					@mxx_sources_root = a_root
				end

				return old_root
			end

			#   .
			def	screen_mode( a_mode )
				if SCREEN_CONSOLE == a_mode || SCREEN_WINDOW == a_mode
					@mxx_screen_mode = a_mode
				else
					raise Mxx_ru::Unsupported_mode_ex.new(
						"Unsupported screen mode: #{a_mode}" )
				end
			end

			#    C-  .
			#
			# _a_file_name_	  .
			#
			# _a_options_	    
			#   .     
			#       .
			def	c_source( a_file_name, a_options = nil )
				@mxx_c_files.push( Cpp_source_file.new(
					create_full_src_file_name( a_file_name ), a_options ) )
			end

			#    C-  .
			#
			# _a_file_name_	  .
			#
			# _a_options_	    
			#   .     
			#       .
			def	cpp_source( a_file_name, a_options = nil )
				@mxx_cpp_files.push( Cpp_source_file.new(
					create_full_src_file_name( a_file_name ), a_options ) )
			end

			#      mswin.
			#
			# _a_rc_file_
			#  rc-.
			#
			# _a_depends_
			#   .
			#   Array  String.
			def	mswin_rc_file(
				a_rc_file,
				a_depends =nil )

				@mxx_mswin_rc_file = Mswin_rc_file.new( a_rc_file, a_depends )
			end

			#       mswin.
			def	mswin_res_file( a_res_file )
				@mxx_mswin_res_file = String.new( a_res_file )
			end

			#       .
			def obj_file( file_name )
				@mxx_obj_files.push( String.new( file_name ) )
			end

			#   .
			#  a_path   nil,    
			#   ,     .
			def	lib( a_library, a_path = nil )
				mxx_add_required_lib( a_library )
				if a_path
					mxx_add_required_lib_path( a_path )
				end
			end

			#    .
			def implib_path( name )
				@mxx_implib_path = name.clone
			end

			#    .
			def	optimization( a_mode )
				if OPTIM_SIZE != a_mode && OPTIM_SPEED != a_mode
					raise Mxx_ru::Invalid_value_ex.new(
						"Unknown optimization mode: '#{a_mode}'" )
				else
					@mxx_optimization = a_mode.clone
				end
			end

			#   .
			def	build
				if nil == @mxx_last_build_result

					if !@mxx_all_options_defined
						# ,  toolset   
						#     .
						Mxx_ru::Cpp::toolset.setup_mandatory_options( self )
						@mxx_all_options_defined = true
					end

					Mxx_ru::Cpp::Mode.instance.before_subprj_processing
					#    
					#     dry-run.
					required_prjs_state = build_required_projects
					Mxx_ru::Cpp::Mode.instance.after_subprj_processing

					run_generators
					if !Mxx_ru::Cpp::Mode.instance.is_no_depends_analyzer
						run_depends_analyzer
					end
					objs_state = compile_sources

					res_state = compile_resources

					link_state = link_target(
						required_prjs_state,
						objs_state, res_state )

					#      .
					@mxx_last_build_result = link_state
				end

				return @mxx_last_build_result
			end

			#   .
			def clean
				if nil == @mxx_last_build_result

					if !@mxx_all_options_defined
						# ,  toolset   
						#     .
						Mxx_ru::Cpp::toolset.setup_mandatory_options( self )
						@mxx_all_options_defined = true
					end

					Mxx_ru::Cpp::Mode.instance.before_subprj_processing
					#   
					#     dry-run.
					clean_required_prjs
					Mxx_ru::Cpp::Mode.instance.after_subprj_processing

					clean_generators
					clean_objs
					clean_resources
					clean_target

					@mxx_last_build_result = Mxx_ru::Target_state.new(
						Mxx_ru::Target_state::ABSENT )
				end
			end

			#   .
			#      .
			#  ,   Mxx_ru::Cpp::Target_type.
			def	target_type
				raise Abstract_method_ex.new( "Mxx_ru::Cpp::Target::target_type" )
			end

			#       
			#   mxx_target_root
			def	create_full_target_file_name( a_file_name )
				if "" == mxx_target_root
					return a_file_name
				else
					return File.join( [ mxx_target_root, a_file_name ] )
				end
			end

			#       
			#   mxx_sources_root
			def	create_full_src_file_name( a_file_name )
				if "" == mxx_sources_root
					return a_file_name
				else
					return File.join( [ mxx_sources_root, a_file_name ] )
				end
			end

		protected
			#     .
			def	run_generators
				mxx_generators.each { |g| g.build( self ) }
			end

			#   C++ .
			def	run_depends_analyzer
				analyzer = Analyzer::Cpp_analyzer.instance

				#    .
				sources = Array.new
				mxx_c_files.each { |s| sources << s.name }
				mxx_cpp_files.each { |s| sources << s.name }
				analyzer.add_sources( sources )

				analyzer.add_include_paths( mxx_all_include_paths )

				#  .
				analyzer.analyze

				#   .
				mxx_c_files.each { |s|
					depends = analyzer.get_depends_for( s.name )
					s.depends << depends
					s.depends.flatten!
				}

				mxx_cpp_files.each { |s|
					depends = analyzer.get_depends_for( s.name )
					s.depends << depends
					s.depends.flatten!
				}
			end

			#    .
			#  Mxx_ru::Target_state.
			def	compile_sources
				return Mxx_ru::Cpp::toolset().make_objs( self )
			end

			#    .
			#   .
			def	compile_resources
				#   ,     
				#       
				#  .
				if @mxx_mswin_rc_file
					return Mxx_ru::Cpp::toolset().make_mswin_res( self )
				end

				return Mxx_ru::Target_state.new( Target_state::EXISTS )
			end

			#   .
			#
			# _required_prjs_state_
			#     .
			#   Mxx_ru::Target_state.
			#
			# _objs_state_
			#     .
			#   Mxx_ru::Target_state.
			#
			# _res_state_
			#    .
			#   Mxx_ru::Target_state.
			def	link_target(
				required_prjs_state,
				objs_state,
				res_state )

				state = nil
				toolset = Mxx_ru::Cpp::toolset
				if nil == target_type
					raise Mxx_ru::Unsupported_target_type_ex.new(
						self.class.name, "nil" )
				elsif target_type.name == Lib_target_type::TYPE
					state = toolset.make_lib( self )
				elsif target_type.name == Dll_target_type::TYPE
					state = toolset.make_dll( self )
				elsif target_type.name == Exe_target_type::TYPE
					state = toolset.make_exe( self )
				else
					raise Mxx_ru::Unsupported_target_type_ex.new(
						self.class.name, target_type.name )
				end

				return state
			end

			def	clean_generators
				mxx_generators.each { |g| g.clean( self ) }
			end

			def	clean_objs
				Mxx_ru::Cpp::toolset().clean_objs( self )
			end

			def	clean_resources
				if @mxx_mswin_rc_file
					Mxx_ru::Cpp::toolset().clean_mswin_res( self )
				end
			end

			def	clean_target
				toolset = Mxx_ru::Cpp::toolset
				if nil == target_type
					raise Mxx_ru::Unsupported_target_type_ex.new(
						self.class.name, "nil" )
				elsif target_type.name == Lib_target_type::TYPE
					state = toolset.clean_lib( self )
				elsif target_type.name == Dll_target_type::TYPE
					state = toolset.clean_dll( self )
				elsif target_type.name == Exe_target_type::TYPE
					state = toolset.clean_exe( self )
				else
					raise Mxx_ru::Unsupported_target_type_ex.new(
						self.class.name, target_type.name )
				end
			end

			#    ,   
			# ,    .
			#
			# _a_option_desc_
			#     .
			def	add_spreadable_option(
				a_option_desc )

				globals = @@mxx_globals[ a_option_desc ]
				if nil == globals
					@@mxx_globals[ a_option_desc ] = globals = Array.new
				end

				locals = Spreadable_option.new( a_option_desc )
				@mxx_locals[ a_option_desc ] = locals

				meth_global = "global_#{a_option_desc}"
				meth_local = a_option_desc
				meth_all = "mxx_all_#{a_option_desc}s"
				meth_upspreads = "mxx_upspread_#{a_option_desc}s"

				@mxx_missing_actions[ meth_global ] =
					Global_option_inserter.new( a_option_desc, globals )
				@mxx_missing_actions[ meth_local ] =
					Spreadable_option_inserter.new( locals )
				@mxx_missing_actions[ meth_all ] =
					Spreadable_option_all_extractor.new( locals,
						globals, self, meth_upspreads )
				@mxx_missing_actions[ meth_upspreads ] =
					Spreadable_option_upspread_extractor.new(
						locals, self, meth_upspreads )
			end

			def add_unique_to( what, to )
				if !to.include?( what )
					to.push( what )
				end
			end

		end # class Target

		#    ,     lib,  dll.
		#
		#         
		# -target.      
		# :
		# *    ,      
		#  .      
		# as_lib;
		# *  ,       
		# .       as_dll.
		# :
		#	class	Prj < Mxx_ru::Cpp::Lib_or_dll_target
		#		TAG = "threads_1/prj.rb"
		#		def	initialize( a_alias = TAG )
		#			super( a_alias, TAG )
		#
		#			target( "threads.1.3" )
		#
		#			threading_mode( Mxx_ru::Cpp::THREADING_MULTI )
		#
		#			init_dll_block(
		#				Proc.new {
		#					rtl_mode( Mxx_ru::Cpp::RTL_SHARED )
		#					implib_path( "lib" )
		#					define( "THREADS_1__DLL", OPT_UPSPREAD )
		#				})
		#
		#			cpp_source( "threads.cpp" )
		#			cpp_source( "micro_time.cpp" )
		#
		#			if toolset.tag( "target_os" ) == "mswin"
		#				define( "__WIN32__" )
		#				cpp_source( "win32/os_thread.cpp" )
		#			elsif toolset.tag( "target_os" ) == "mswin"
		#				cpp_source( "posix/os_thread.cpp" )
		#			end
		#
		#			define( "THREADS_1__PRJ" )
		#		end
		#	end
		#
		#      ,  
		#     
		#   .
		#
		# ==    as_lib, as_dll
		#
		#      , 
		#  Lib_or_dll_target,      
		#    as_lib  as_dll    required_prj:
		#
		#	class	My_prj < Mxx_ru::Cpp::Exe_target
		#		def	initialize( a_alias = "my_prj.rb" )
		#			super( a_alias )
		#			required_prj( "threads_1/prj.rb" ).as_dll( self )
		#		end
		#	end
		#
		#    ,      
		#    threads_1/prj.rb,    
		#     as_dll.     , ..
		# as_dll        .
		#
		# ==   -
		#
		#    ,    
		# lib,  dll,       ,
		#      as_lib  as_dll  
		# :
		#
		#	class	Prj < Mxx_ru::Cpp::Lib_or_dll_target
		#		def	initialize( a_alias = "threads_1/prj.rb" )
		#			super( a_alias, "threads_1" )
		#			...
		#		end
		#	end
		#	class	Lib < Prj
		#		def super( a_alias = "threads_1/lib.rb" )
		#			super( a_alias )
		#			as_lib
		#		end
		#	end
		#	class	Dll < Prj
		#		def super( a_alias = "threads_1/dll.rb" )
		#			super( a_alias )
		#			as_dll
		#		end
		#	end
		#
		#       :
		#	class	My_prj < Mxx_ru::Cpp::Exe_target
		#		def	initialize( a_alias = "my_prj.rb" )
		#			super( a_alias )
		#			required_prj( "threads_1/dll.rb" )
		#		end
		#	end
		#
		class	Lib_or_dll_target < Target
			#  ,  ,     .
			class	Type_not_defined_ex < Mxx_ru::Ex
				# _a_prj_alias_	 ,    
				#   .
				def	initialize( a_prj_alias )
					super( "#{a_prj_alias}: type of target not defined yet" )
				end
			end

			#  ,  ,   
			#  .
			class	Type_already_defined_ex < Mxx_ru::Ex
				# _a_prj_tag_	 ,   
				# .
				#
				# _a_type_	   .
				def	initialize( a_prj_alias, a_type )
					super( "#{a_prj_alias}: type of target already defined " +
						"as #{a_type}" )
				end
			end

			#       .
			#    , 
			#           
			# lib,   dll.
			@@mxx_tags = Hash.new

			# .
			#
			# _a_prj_alias_	 .
			#
			# _a_prj_tag_	 ,   
			# .      
			#   :
			#	class	Prj < Mxx_ru::Cpp::Lib_or_dll_target ... end
			#	class	Lib < Prj ... end
			#	class	Dll < Prj ... end
			#	class	My_exe < Mxx_ru::Cpp::Exe_target
			#		def	initialize( a_alias = "my_exe.rb" )
			#			super( a_alias )
			#			required_prj( "threads_1/dll.rb" )
			#		end
			#	end
			#	class	My_dll < Mxx_ru::Cpp::Dll_target
			#		def	initialize( a_alias = "my_dll.rb" )
			#			super( a_alias )
			#			required_prj( "threads_1/lib.rb" )
			#		end
			#	end
			#  My_exe  My_dll     ,
			#     threads_1/dll.rb  threads_1/lib.rb.
			def initialize( a_prj_alias, a_prj_tag )
				super( a_prj_alias )

				@mxx_prj_tag = a_prj_tag

				# ,   ,    lib.
				@mxx_init_lib_block = nil
				# ,   ,    dll.
				@mxx_init_dll_block = nil
				#        as_lib,
				#   as_dll.
				@mxx_target_type = nil

			end

			#      lib.
			def as_lib
				if ensure_type_not_set_yet( Lib_target_type::TYPE )
					@mxx_target_type = Lib_target_type.new
					if nil != @mxx_init_lib_block
						@mxx_init_lib_block.call
					end
				end
			end

			#      dll.
			def as_dll
				if ensure_type_not_set_yet( Dll_target_type::TYPE )
					@mxx_target_type = Dll_target_type.new
					if nil != @mxx_init_dll_block
						@mxx_init_dll_block.call
					end
				end
			end

			#    mxx_target_type.
			def target_type
				if nil == @mxx_target_type
					raise Type_not_defined_ex.new( prj_alias )
				end

				return @mxx_target_type
			end

		protected
			#     lib.
			#       as_lib.
			def	init_lib_block( a_block )
				@mxx_init_lib_block = a_block
			end

			#     dll.
			#       as_dll.
			def	init_dll_block( a_block )
				@mxx_init_dll_block = a_block
			end

		private
			#  -     , 
			#  .     
			#  ,   false.     ,
			#   true.
			#
			# _a_type_name_    .
			def	ensure_type_not_set_yet( a_type_name )
				if @@mxx_tags.key?( @mxx_prj_tag ) &&
					@@mxx_tags[ @mxx_prj_tag ] != a_type_name
					raise	Type_already_defined_ex.new(
						@mxx_prj_tag, a_type_name )
				end

				@@mxx_tags[ @mxx_prj_tag ] = a_type_name

				return ( nil == @mxx_target_type )
			end
		end # class Lib_or_dll_target

		#    lib.
		class	Lib_target < Lib_or_dll_target
			def initialize( a_prj_alias, &block )
				super( a_prj_alias, a_prj_alias )

				as_lib

				instance_eval( &block ) if block
			end
		end

		#    dll.
		class	Dll_target < Lib_or_dll_target
			def initialize( a_prj_alias, &block )
				super( a_prj_alias, a_prj_alias )

				as_dll

				instance_eval( &block ) if block
			end
		end

		#    exe.
		class	Exe_target < Target

			def	initialize( a_alias, &block )
				super( a_alias )

				instance_eval( &block ) if block
			end

			def target_type
				return Exe_target_type.new
			end
		end
	end # module Cpp
end # module Mxx_ru
